import os
import json
from .clock import Clock

class ObjectStore:
    def __init__(self, clock: Clock, path: str):
        self.clock = clock
        self.data_path = path
        
    def save(self, key: str, value: str | list | dict, namespace=None):
        path = self.data_path
        if isinstance(value, list) or isinstance(value, dict):
            value = json.dumps(value, indent=2)
        if namespace is not None:
            path = os.path.join(path, namespace)
        os.makedirs(path, exist_ok=True)
        file_path = os.path.join(path, key + ".json")
        with open(file_path, "w") as f:
            f.write(value)
        
    def load(self, key: str, namespace=None):
        path = self.data_path
        if namespace is not None:
            path = os.path.join(path, namespace)
        file_path = os.path.join(path, key + ".json")
        if not os.path.exists(file_path):
            return None
        with open(file_path, "r") as f:
            return f.read()
        
    def load_all(self, namespace=None):
        path = self.data_path
        if namespace is not None:
            path = os.path.join(path, namespace)
        data = []
        if not os.path.exists(path):
            return data
        for file in os.listdir(path):
            obj = self.load_obj(file.replace(".json", ""), None, namespace)
            data.append(obj)
        return data
        
    def load_obj(self, key: str, Cls=None, namespace=None):
        data = self.load(key, namespace)
        if data is None:
            return None
        if Cls is None:
            return json.loads(data)
        
        return Cls.model_validate_json(data)




def _reflection_with_llm_as_summary_dict(sender, recipient, summary_args):  
    summary_dict = {}
    for key, prompt in summary_args.items():
        if not isinstance(prompt, str):
            continue
        msg_list = recipient.chat_messages_for_summary(sender)
        agent = sender if recipient is None else recipient
        role = summary_args.get("summary_role", None)
        if role and not isinstance(role, str):
            raise ValueError("The summary_role in summary_arg must be a string.")

        summary = sender._reflection_with_llm(
            prompt, msg_list, llm_agent=agent, cache=summary_args.get("cache"), role=role
        )
        summary_dict[key] = summary
    return summary_dict



from autogen.agentchat.contrib.agent_builder import AgentBuilder
import json


class University:
    def __init__(self, llm_config):
        self.agents = []
        self.llm_config = llm_config

        #TODO: change the hard code here
        model = (
            llm_config.get('config_list', [{}])[0].get('model', 'gpt-4o-mini')
        )
        self.builder_agent_model = model
        self.agent_model = model

    def add_agent(self, agent):
        self.agents.append(agent)

    def list_agents(self):
        for agent in self.agents:
            print(f"University Agent: {agent.name}")

    def build_agents(self, message, ):
        config_path = 'agent_config.json'
        config_list = self.llm_config['config_list']

        with open(config_path, 'w') as f:
            json.dump(config_list, f)

        builder = AgentBuilder(
            config_file_or_env=config_path,
            builder_model=self.builder_agent_model,
            agent_model=self.agent_model
        )

        # Define the building task for investment agents
        # TODO change the hard code here
        building_task = """Create specialized agents for investment analysis and execution:
                1. A planner to coordinate the overall investment strategy and approve plans
                2. A data analyst to analyze market data and provide investment recommendations
                3. An investment executor to implement approved investment plans using provided tools
                The agents should work together to analyze stock data, create investment plans, and execute trades."""

        # Build the agents using AgentBuilder
        agent_list, _ = builder.build(building_task, self.llm_config, coding=True)

        return agent_list



from .app import App
from .clock import Clock
from .object_store import ObjectStore
from typing import Annotated
from pydantic import BaseModel

class JobPosting(BaseModel):
    skill: str
    description: str
    duties: str

class JobBoard(App):
    clock: Clock
    store: ObjectStore
    job_board: list[JobPosting] = []
    
    def __init__(self, clock: Clock, store: ObjectStore):
        super().__init__(clock, store)
        self.job_board = []
        
        
    def add_posting(self, posting: JobPosting) -> str:
        self.job_board.append(posting)
        self.on_change()
        return "Posting added"
    
    
    def get_postings(self) -> list:
        return [j.model_dump() for j in self.job_board]
    
    def get_posting_markdown(self, posting: JobPosting) -> str:
        return f"Skill: {posting.skill}\nDescription: {posting.description}\nDuties: {posting.duties}"
    
    def get_postings_as_text(self) -> str:
        postings = self.get_postings()
        table = "| Skill | Description | Duties \n"
        table += "| --- | --- | --- | \n"
        for posting in postings:
            table += f"| {posting['skill']} | {posting['description']} | {posting['duties']} \n"
        return table
    
    def clear_postings(self) -> str:
        self.job_board = []
        self.on_change()
        return "Postings cleared"
    
    def get_functions(self):
        def add_job_posting(
            skill: Annotated[str, "The skill required for the job posting"],
            description: Annotated[str, "A description of the job posting"],
            duties: Annotated[str, "The duties of the job posting"]
            ) -> str:
            return self.add_posting(JobPosting(skill=skill, description=description, duties=duties))
        
        def get_job_postings() -> list:
            return self.get_postings()
        
        def clear_job_postings() -> str:
            return self.clear_postings()
        
        return[
            (add_job_posting, "Adds a new skill posting to the job board"),
            (get_job_postings, "Returns the current job postings"),
            (clear_job_postings, "Clears all job postings")
        ]
    
    def on_change(self):
        self.store.save("job_board", {
            "postings": self.get_postings()
        }, namespace="job_board")


from pydantic import BaseModel
from datetime import datetime
from .apps.app import App
from .apps.clock import Clock
from .apps.task_manager import TaskManager
from .apps.constitution import Constitution
from autogen import AssistantAgent
from .apps.object_store import ObjectStore
from .apps.budget import BudgetApp
from .apps.job_board import JobBoard

class SystemConfig(BaseModel):
    clock_start: datetime
    data_path: str

class System:
    config: SystemConfig
    clock: Clock
    apps: dict[str, App]
    task_manager: TaskManager
    constitution: Constitution
    
    def __init__(self, config=SystemConfig | dict):
        if isinstance(config, dict):
            config = SystemConfig(**config)
        
        self.config = config
        
        self.clock = Clock(config.clock_start)
        self.store = ObjectStore(self.clock, config.data_path)
        self.task_manager = TaskManager(self.clock, self.store)
        self.constitution = Constitution(self.clock, self.store)
        self.apps = {
            "task_manager": self.task_manager,
            "constitution": self.constitution,
            "budget": BudgetApp(self.clock, self.store),
            "job_board": JobBoard(self.clock, self.store)
        }
        
    def get_executor(self):
        return AssistantAgent(
            name="Executor",
            description="Use this agent to execute calls to tools",
            system_message="You ONLY execute tools calls and returns the responses.",
        )


from autogen import ConversableAgent

class TeamMember:
    name: str
    ag_agent: ConversableAgent
    apps: list[str]
    
    @staticmethod
    def from_autogen_agent(agent: ConversableAgent, apps: list[str] = []):
        member = TeamMember()
        member.name = agent.name
        member.ag_agent = agent
        member.apps = apps
        
    def get_apps(self):
        return []
    
    def get_agent(self, llm_config = None) -> tuple:
        return self.ag_agent


from .clock import Clock
from .object_store import ObjectStore
from autogen import AssistantAgent

class App:
    clock: Clock
    store: ObjectStore
    def __init__(self, clock: Clock, store: ObjectStore):
        self.clock = clock
        self.store = store
        
    def train_agent(self, assistant: AssistantAgent):
        functions = self.get_functions()
        for function, description in functions:
            name = self.get_full_method_name(function.__name__)
            assistant.register_for_llm(name=name, description=description)(function)
    
    
    def has_trained(self, agent: AssistantAgent):
        functions = self.get_functions()
        for function, _ in functions:
            name = self.get_full_method_name(function.__name__)
            if name in agent.function_map:
                return True
        
    
    def get_functions(self):
        raise NotImplementedError()
    
    def get_full_method_name(self, method_name: str) -> str:
        return self.__class__.__name__ + "_" + method_name
    
    def register_executor(self, executor: AssistantAgent):
        functions = self.get_functions()
        for function, _ in functions:
            name = self.get_full_method_name(function.__name__)
            # Register the function to execute
            executor.register_for_execution(name=name)(function)
        
    
    
    


from autogen import AssistantAgent
from .ceo import CEO
from .treasurer import Treasurer
from .hr_manager import HRManager

class Govern:
    def __init__(self, llm_config):
        self.llm_config = llm_config
        self.members = [
            CEO(),
            Treasurer(),
            HRManager()
        ]
        
    def get_members(self) -> list[AssistantAgent]:
        return [ member for member in self.members ]





from datetime import datetime, timedelta

class Clock:
    def __init__(self, time: datetime):
        self.time = time

    def tick(self):
        self.time += timedelta(minutes=10)

    def now(self):
        return self.time
    
    def set_time(self, time: datetime):
        self.time = time
    
    def advance(self, hours: int = None, days: int = None):
        if hours is None and days is None:
            raise ValueError("You must provide either hours or days")
        if hours is not None:
            self.time += timedelta(hours=hours)
        if days is not None:
            self.time += timedelta(days=days)
        return self.time
    
    def __str__(self):
        return f"{self.time} o'clock"
   


from autogen import AssistantAgent
from wandagents.economy.system import System
from wandagents.team.member import TeamMember

MAIN_PROMPT = """
You are the CEO of the team. You are responsible for its overall success.
You are a master in strategy, and is able to see the big picture.
You are obsessed with the team's goals and you are always looking for ways to improve the team's performance.
You design teams and processes to achieve the team's goals.
You can provide guidance and direction to the team. and delegate tasks to the team members.
You write inpiring mission statements and you are able to communicate the team's vision to the team members.
You will participate in worksessions, you don't need to introduce yourself, and when the work session task is achieved you write 'WORK SESSION COMPLETED'.
"""

class CEO(TeamMember):
    
    def __init__(self):
        self.name = "CEO"
        
    def get_apps(self):
        return ["task_manager", "constitution"]
    
    def get_agent(self, llm_config) -> tuple:
        assistant = AssistantAgent(
            name="CEO",
            description="Chief Executive Officer",
            system_message=MAIN_PROMPT,
            code_execution_config=False,
            human_input_mode="NEVER",
            llm_config=llm_config,
        )
        
        return assistant
        
        
        
        
        


import sys
import os
import json
from pydantic import BaseModel
from datetime import datetime

class Config(BaseModel):
    clock_start: datetime = datetime.now()
    data_path: str
    model: str

def load_config() -> Config:
    config_flag = [arg for arg in sys.argv if arg.startswith("--config=")]
    if config_flag:
        config_file = config_flag[0].split("=")[1]
    elif os.getenv("CONFIG_FILE"):
        config_file = os.getenv("CONFIG_FILE")
    else:
        config_file = "./config.json"
    
    with open(config_file, "r") as f:
        config = json.load(f)
    return Config(**config)


from autogen import AssistantAgent
from wandagents.economy.system import System
from wandagents.team.member import TeamMember

MAIN_PROMPT = """
You are the HR Manager of the team. You are responsible for ensuring the team has the skills needed while ensuring compliance to the mission and budget.
Hiring and firing of team members is your responsibility.
Each tem member has a cost, so keep that in consideraton.    
You will participate in work sessions, you don't need to introduce yourself, and when the work session task is achieved you write 'WORK SESSION COMPLETED'.
"""

class HRManager(TeamMember):
    
    def __init__(self):
        self.name = "HRManager"
        
    def get_apps(self):
        return ["task_manager", "job_board"]
    
    def get_agent(self, llm_config) -> tuple:
        assistant = AssistantAgent(
            name=self.name,
            description="This is the HR Manager of the team. Responsible for managing the team's human resources",
            system_message=MAIN_PROMPT,
            code_execution_config=False,
            human_input_mode="NEVER",
            llm_config=llm_config,
        )
        
        return assistant
        
        
        
        
        


from pydantic import BaseModel
from typing import Annotated
from .clock import Clock
from autogen import AssistantAgent
import pandas as pd
from .app import App
from .object_store import ObjectStore

class BudgetApp(App):
    funds: float = 0
    accounts: dict = {}
    history: list[dict] = []
    clock: Clock
    store: ObjectStore
    
    def __init__(self, clock: Clock, store: ObjectStore):
        self.clock = clock
        self.store = store
    
    def set_available_funds(self, amount: float) -> str:
        self.accounts["Not Allocated"] = amount
        self.on_change()
        return "Allocated %.2f to Not Allocated" % amount
    
    def allocate(self, 
                 category: Annotated[str, "The category to which allocate funds" ], 
                 amount: Annotated[float, "The amount to allocate"]) -> str:
        
        self.accounts["Not Allocated"] > amount
        self.accounts["Not Allocated"] -= amount
        if self.accounts.get(category):
            self.accounts[category] += amount
        else:
            self.accounts[category] = amount
        self.on_change()
        return "Allocated"

    def get_avaliable_funds(self, 
                category: Annotated[str, "the category to get the available funds. If not provided it will return all funds available" ] = None) -> float:
        if category is None:
            return self.get_total_avaliable_funds()
        return self.accounts[category]
    
    def get_allocations(self) -> dict:
        return self.accounts
    
    def get_current_budget(self) -> str:
        if len(self.accounts) == 0:
            return ""
        allocations = "Current Budget:"
        for category, amount in self.accounts.items():
            allocations += f"{category}: {amount}\n"
        return allocations
        
    def get_total_avaliable_funds(self) -> float:
        return sum(self.accounts.values())
    
    
    
    def spend(self, amount: float,  category: str = None) -> str:
        provisioned = self.accounts.get(category)
        if category not in self.accounts:
            self.accounts["Not Allocated"] -= amount
        else:
            self.accounts[category] -= amount
        
        self.history.append({
            "date": self.clock.now().isoformat(),
            "category": category,
            "amount": amount * -1,
            "type": "spend"
        })
        self.on_change()
        if provisioned is None:
            return "Spent %.2f from Not Allocated as %s was not provisioned" % (amount,category)
        return "Spent %.2f from %s" % (amount, category)
        
    def deposit(self, category: str, amount: float) -> str:
        if category not in self.accounts:
            self.accounts[category] = 0
        self.accounts[category] += amount    
        self.history.append({
            "date": self.clock.now().isoformat(),
            "category": category,
            "amount": amount,
            "type": "deposit"
        })
        self.on_change()
        return "Deposited %.2f to %s" % (amount, category)
    
    def on_change(self):
        self.store.save("budget", {
            "funds": self.funds,
            "accounts": self.accounts,
            "history": self.history
        }, namespace="budget")
    
    def get_cashflow(self, 
                     start_date: Annotated[str, "The start date"], 
                     end_date: Annotated[str, "The end date"] ) -> str:
        df = pd.DataFrame(self.history)
        df = df[(df["date"] >= start_date) & (df["date"] <= end_date)]
        return df.groupby("category").sum("amount").to_markdown()
    
    def serialize(self) -> str:
        return self.model_dump_json()
    
    def get_functions(self):
        def set_available_funds(amount: float) -> str:
            return self.set_available_funds(amount)
        
        def allocate(category: str, amount: float) -> str:
            return self.allocate(category, amount)
        
        def get_avaliable_funds(category: str) -> float:
            return self.get_avaliable_funds(category)
        
        def get_total_avaliable_funds() -> float:
            return self.get_total_avaliable_funds()
        
        def spend(category: str, amount: float) -> str:
            return self.spend(amount, category)
        
        def deposit(category: str, amount: float) -> str:
            return self.deposit(category, amount)
        
        def get_cashflow(start_date: str, end_date: str) -> str:
            return self.get_cashflow(start_date, end_date)
        
        def get_allocations() -> dict:
            return self.get_allocations()
        
        return [
            (set_available_funds, "Use this tool to set the available funds when starting an account and there were no funds. The amount will be allocated to the `Not Allocated` category"),
            (allocate, "Use this tool to allocate funds to a category"),
            (get_avaliable_funds, "Use this tool to get the avaliable funds for a category"),
            (get_total_avaliable_funds, "Use this tool to get the total avaliable funds"),
            (spend, "Use this tool to spend funds from a category"),
            (deposit, "Use this tool to deposit funds to a category"),
            (get_cashflow, "Use this tool to get the cashflow for a period of time"),
            (get_allocations, "Use this tool to get the allocations for the budget")
        ]
        


import json
from .app import App
from .clock import Clock
from .object_store import ObjectStore
from autogen import AssistantAgent
from pydantic import BaseModel
from enum import Enum
from typing import Annotated


class Task(BaseModel):
    assignee: Annotated[str, "The person assigned to the task"]
    name: Annotated[str, "The name of the task"]
    description: Annotated[str, "The description of the task"]
    status: Annotated[str, "The status of the task it only accepts 'pending' or 'done'"]
    type: Annotated[str, "The type of the task. Discussion is a task that will benefit from a meeting. Work is a task that can be done by the assignee and reported to the team"]
    participants: Annotated[list[str], "The additional participants of the task. For discussion tasks, these will participate in the discussion"] = None
    schedule_for: Annotated[str, "The date the task is scheduled for. If set to None it will be executed ASAP"] = None

class TaskManager(App):
    tasks: list[Task] = []
    
    def __init__(self, clock: Clock, store: ObjectStore):
        super().__init__(clock, store)
        self.tasks = []
    
    def create_task(self, task: Annotated[Task, "The task to be created"]) -> str:
        task.status = "pending"
        self.tasks.append(task)
        self.on_change(task)
        return f"Task {task.name} created"
    
    def get_tasks(self) -> list:
        return [t for t in self.tasks if t.status not in ["done", "completed"] ]
    
    def get_task(self, name: str) -> str:
        for task in self.tasks:
            if task.name == name:
                return self.task_to_mkd(task)
        return None    
    
    def task_to_mkd(self, task: Task) -> str:
        text = f"Task: {task.name}\n"
        text += f"Description: {task.description}\n"
        text += f"Assignee: {task.assignee}\n"
        text += f"Status: {task.status}\n"
        text += f"Scheduled For: {task.schedule_for}\n"
        return text
        
    
    def complete_task(self, name: str) -> str:
        for task in self.tasks:
            if task.name == name:
                task.status = "done"
                self.on_change(task)
                return "Task completed"
        return "Task not found"
    
    def delete_task(self, name: str) -> str:
        for task in self.tasks:
            if task.name == name:
                self.tasks.remove(task)
                self.on_change(task)
                return "Task deleted"
        return "Task not found"
    
    def get_tasks_as_text(self) -> str:
        """Returns the tasks as a markdown table"""
        tasks = self.get_tasks()
        table = "| Name | Description | Assignee | Status \n"
        table += "| --- | --- | --- | --- | \n"
        for task in tasks:
            table += f"| {task.name} | {task.description} | {task.assignee} | {task.status} \n"
        return table

    def to_json(self):
        return  [json.loads(t.model_dump_json()) for t in self.tasks]
    
    def on_change(self, task: Task = None):
        self.store.save("tasks", self.to_json(), namespace="task_manager")
    
    def get_functions(self):
        def create_task(task: Task) -> str:
            return self.create_task(task)
        
        def get_tasks() -> str:
            return self.get_tasks_as_text()
        
        def get_task(name: str) -> dict:
            return self.get_task(name)
        
        def complete_task(name: str) -> str:
            return self.complete_task(name)
        
        def delete_task(name: str) -> str:
            return self.delete_task(name)
        
        return [
            (create_task, "Use this tool to create a task"), 
            (get_tasks, "Use this tool to get the tasks"), 
            (get_task, "Use this tool to get a task"), 
            (complete_task, "Use this tool to complete a task"), 
            (delete_task, "Use this tool to delete a task")
        ]
    
  


from .app import App
from .clock import Clock
from .object_store import ObjectStore
from typing import Annotated

class Constitution(App):
    clock: Clock
    store: ObjectStore
    def __init__(self, clock: Clock, store: ObjectStore):
        self.clock = clock
        self.store = store
        self.mission = ""
        self.plan = ""
        self.initial_budget = 0
    
    def set_mission(self, mission: str) -> str:
        self.mission = mission
        self.on_change()
        return "Mission set"
    
    def set_strategic_plan(self, plan: Annotated[str, "A concise strategic plan for the team"]) -> str:
        self.plan = plan
        self.on_change()
        return "Plan set"
    
    def set_initial_budget(self, budget: float) -> str:
        self.initial_budget = budget
        self.on_change()
        return "Initial budget set"
    
    def get_initial_budget(self) -> float:
        return self.initial_budget
    
    def get_strategic_plan(self) -> str:   
        return self.plan
    
    def get_mission(self) -> str:
        return self.mission
    
    def on_change(self):
        self.store.save("constitution", {
            "mission": self.mission,
            "plan": self.plan,
            "initial_budget": self.initial_budget
        }, namespace="constitution")
        pass
    
    def get_constitution(self) -> str:
        if not self.mission:
            return ""
        constitution = "# Team's Constitution\n"
        constitution += f"## Mission\n{self.mission}\n"
        constitution += f"## Strategic Plan\n{self.plan}\n"
        constitution += f"## Initial Budget\n{self.initial_budget}\n"
        return constitution
        
        
    
    def get_functions(self):
        def set_mission(mission: str) -> str:
            return self.set_mission(mission)
        
        def get_mission() -> str:
            return self.get_mission()
        
        def set_strategic_plan(plan: str) -> str:
            return self.set_strategic_plan(plan)
        
        def set_initial_budget(budget: float) -> str:
            return self.set_initial_budget(budget)
        
        def get_initial_budget() -> float:
            return self.get_initial_budget()
        
        def get_strategic_plan() -> str:
            return self.get_strategic_plan()
        
        return [
            (set_mission, "Use this tool to set the mission of the team"),
            (get_mission, "Use this tool to get the mission of the team"),
            (set_strategic_plan, "Use this tool to set the strategic plan of the team"),
            (set_initial_budget, "Use this tool to set the initial budget of the team"),
            (get_initial_budget, "Use this tool to get the initial budget of the team"),
            (get_strategic_plan, "Use this tool to get the strategic plan of the team")
        ]


from autogen import AssistantAgent
from wandagents.economy.system import System
from wandagents.team.member import TeamMember

MAIN_PROMPT = """
You are the Treasurer of the team. You are responsible for budgeting and controling spent.
You are obsessed with ensuring the budget is well mantianed and consiliated with reality.
If given a task to estabilish the team you set the initial budget and allocations
You will participate in work sessions, you don't need to introduce yourself, and when the work session task is achieved you write 'WORK SESSION COMPLETED'.
"""

class Treasurer(TeamMember):
    
    def __init__(self):
        self.name = "Treasurer"
        
    def get_apps(self):
        return ["task_manager", "budget"]
    
    def get_agent(self, llm_config) -> tuple:
        assistant = AssistantAgent(
            name=self.name,
            description="This is the Treasurer, reachout to him if you need financial resources",
            system_message=MAIN_PROMPT,
            code_execution_config=False,
            human_input_mode="NEVER",
            llm_config=llm_config,
        )
        
        return assistant
        
        
        
        
        


from wandagents.gov.govern import Govern
from wandagents.economy.system import System
from .member import TeamMember

class Team:
    govern: Govern
    system: System
    members: TeamMember
    
    def __init__(self, system: System, llm_config):
        self.govern = Govern(llm_config)
        self.system = system
        self.llm_config = llm_config
        self.members = []
        
    def get_team_names(self):
        return [ member.name for member in self.members ]
    
    def get_gov_names(self):
        return [ member.name for member in self.get_govern_members() ]
    
    def get_all_names(self):
        return self.get_team_names() + self.get_gov_names()
    
    def add_member(self, member: TeamMember):
        self.members.append(member)
    
    def get_govern_members(self):
        return self.govern.get_members()
    
    def get_members(self, names: list[str] = None) -> list[TeamMember]:
        govern = self.get_govern_members()
        members = govern + self.members
        if names is not None:
            return [ member for member in members if member.name in names ]
        return govern